"""Provides ability to run jobs locally or on HPC."""

from collections import OrderedDict
import datetime
import fileinput
import importlib
import logging
import os
import shutil

import jade
from jade.common import (
    CONFIG_FILE,
    JOBS_OUTPUT_DIR,
    OUTPUT_DIR,
    RESULTS_FILE,
    HPC_CONFIG_FILE,
)
from jade.enums import JobCompletionStatus, Status, ResourceMonitorType
from jade.events import (
    EVENTS_FILENAME,
    EVENT_NAME_ERROR_LOG,
    StructuredLogEvent,
    EVENT_CATEGORY_ERROR,
    EVENT_CATEGORY_RESOURCE_UTIL,
    EVENT_NAME_BYTES_CONSUMED,
    EVENT_NAME_SUBMIT_STARTED,
    EVENT_NAME_SUBMIT_COMPLETED,
)
from jade.exceptions import InvalidParameter
from jade.extensions.registry import Registry, ExtensionClassType
from jade.hpc.common import HpcType
from jade.hpc.hpc_manager import HpcManager
from jade.hpc.hpc_submitter import HpcSubmitter
from jade.jobs.cluster import Cluster
from jade.jobs.job_configuration_factory import create_config_from_previous_run
from jade.jobs.job_manager_base import JobManagerBase
from jade.jobs.job_runner import JobRunner
from jade.jobs.results_aggregator import ResultsAggregator
from jade.models import SubmitterParams
from jade.models.submission_group import make_submission_group_lookup
from jade.loggers import log_event
from jade.result import serialize_results, ResultsSummary
from jade.utils.repository_info import RepositoryInfo
from jade.utils.subprocess_manager import run_command
from jade.utils.utils import dump_data, get_directory_size_bytes
import jade.version


logger = logging.getLogger(__name__)


class JobSubmitter(JobManagerBase):
    """Submits jobs for execution locally or on an HPC."""

    def __init__(self, config_file, output, is_new):
        """Internal constructor. Callers should use create() or load()."""
        super().__init__(config_file, output)
        self._hpc = None
        self._config_file = config_file
        self._is_new = is_new

    @classmethod
    def create(cls, config_file, params: SubmitterParams, output=OUTPUT_DIR):
        """Creates a new instance.

        Parameters
        ----------
        config_file : JobConfiguration
            configuration for simulation
        params: SubmitterParams
        output : str
            Output directory

        """
        main_file = os.path.join(output, CONFIG_FILE)
        shutil.copyfile(config_file, main_file)
        mgr = cls(main_file, output, True)
        mgr.run_checks(params)
        return mgr

    @classmethod
    def load(cls, output):
        """Loads an instance from an existing directory."""
        return cls(os.path.join(output, CONFIG_FILE), output, False)

    def __repr__(self):
        return f"""num_jobs={self.get_num_jobs()}
results_summary={self.get_results_summmary_report()}"""

    def cancel_jobs(self, cluster):
        """Cancel running and pending jobs."""
        groups = make_submission_group_lookup(cluster.config.submission_groups)
        hpc = HpcManager(groups, self._output)
        for job_id in cluster.job_status.hpc_job_ids:
            hpc.cancel_job(job_id)

        cluster.mark_complete(canceled=True)

    def submit_jobs(self, cluster, force_local=False):
        """Submit simulations. Auto-detect whether the current system is an HPC
        and submit to its queue. Otherwise, run locally.

        Parameters
        ----------
        cluster : Cluster
        force_local : bool
            If on HPC, run jobs through subprocess as if local.

        Returns
        -------
        Status

        """
        if self._is_new:
            logger.info("Submit %s jobs for execution.", self._config.get_num_jobs())
            logger.info("JADE version %s", jade.version.__version__)
            registry = Registry()
            loggers = registry.list_loggers()
            logger.info("Registered modules for logging: %s", ", ".join(loggers))
            self._save_repository_info(registry)

            ResultsAggregator.create(self._output)

            # If an events summary file exists, it is invalid.
            events_file = os.path.join(self._output, EVENTS_FILENAME)
            if os.path.exists(events_file):
                os.remove(events_file)

            event = StructuredLogEvent(
                source="submitter",
                category=EVENT_CATEGORY_RESOURCE_UTIL,
                name=EVENT_NAME_SUBMIT_COMPLETED,
                message="job submission started",
                num_jobs=self.get_num_jobs(),
            )
            log_event(event)
        else:
            self._handle_submission_groups_after_deserialize(cluster)

        result = Status.IN_PROGRESS
        group = self._config.get_default_submission_group()
        groups = make_submission_group_lookup(cluster.config.submission_groups)
        self._hpc = HpcManager(groups, self._output)

        if self._hpc.hpc_type == HpcType.LOCAL or force_local:
            runner = JobRunner(self._config_file, output=self._output)
            num_processes = group.submitter_params.num_processes
            verbose = group.submitter_params.verbose
            result = runner.run_jobs(verbose=verbose, num_processes=num_processes)
            agg = ResultsAggregator.load(self._output)
            agg.process_results()
            is_complete = True
        else:
            is_complete = self._submit_to_hpc(cluster)

        if is_complete:
            result = self._handle_completion(cluster)

        return result

    def _handle_completion(self, cluster):
        result = Status.GOOD
        self._results = ResultsAggregator.list_results(self._output)
        if len(self._results) != self._config.get_num_jobs():
            finished_jobs = {x.name for x in self._results}
            all_jobs = {x.name for x in self._config.iter_jobs()}
            missing_jobs = sorted(all_jobs.difference(finished_jobs))
            logger.error(
                "Error in result totals. num_results=%s total_num_jobs=%s",
                len(self._results),
                self._config.get_num_jobs(),
            )
            logger.error(
                "These jobs did not finish: %s. Check for process crashes or HPC timeouts.",
                missing_jobs,
            )
            result = Status.ERROR
        else:
            missing_jobs = []

        self.write_results_summary(RESULTS_FILE, missing_jobs)
        self._log_error_log_messages(self._output)

        bytes_consumed = get_directory_size_bytes(self._output, recursive=False)
        event = StructuredLogEvent(
            source="submitter",
            category=EVENT_CATEGORY_RESOURCE_UTIL,
            name=EVENT_NAME_BYTES_CONSUMED,
            message="main output directory size",
            bytes_consumed=bytes_consumed,
        )
        log_event(event)

        event = StructuredLogEvent(
            source="submitter",
            category=EVENT_CATEGORY_RESOURCE_UTIL,
            name=EVENT_NAME_SUBMIT_COMPLETED,
            message="job submission completed",
            num_jobs=self.get_num_jobs(),
        )
        log_event(event)

        group = self._config.get_default_submission_group()
        if group.submitter_params.generate_reports:
            self.generate_reports(self._output, group.submitter_params.resource_monitor_type)
        cluster.mark_complete()

        if cluster.config.pipeline_stage_num is not None:
            # The pipeline directory must be the one above this one.
            pipeline_dir = os.path.dirname(self._output)
            next_stage = cluster.config.pipeline_stage_num + 1
            cmd = (
                f"jade pipeline submit-next-stage {pipeline_dir} "
                f"--stage-num={next_stage} "
                f"--return-code={result.value}"
            )
            run_command(cmd)

        return result

    def write_results_summary(self, filename, missing_jobs):
        """Write the results to filename in the output directory."""
        data = OrderedDict()
        data["jade_version"] = jade.version.__version__
        now = datetime.datetime.now()
        data["timestamp"] = now.strftime("%m/%d/%Y %H:%M:%S")
        data["base_directory"] = os.getcwd()
        results = self._build_results(missing_jobs)
        data["results_summary"] = results["summary"]
        data["missing_jobs"] = missing_jobs
        data["results"] = results["results"]

        output_file = os.path.join(self._output, filename)
        dump_data(data, output_file)

        logger.info("Wrote results to %s.", output_file)
        num_successful = results["summary"]["num_successful"]
        num_canceled = results["summary"]["num_canceled"]
        num_failed = results["summary"]["num_failed"]
        num_missing = len(missing_jobs)
        total = num_successful + num_failed + num_missing
        log_func = logger.info if num_successful == total else logger.warning
        log_func(
            "Successful=%s Failed=%s Canceled=%s Missing=%s Total=%s",
            num_successful,
            num_failed,
            num_canceled,
            num_missing,
            total,
        )

        return output_file

    def _build_results(self, missing_jobs):
        num_successful = 0
        num_failed = 0
        num_canceled = 0
        for result in self._results:
            if result.is_successful():
                num_successful += 1
            elif result.is_failed():
                num_failed += 1
            else:
                assert result.is_canceled(), str(result)
                num_canceled += 1

        return {
            "results": serialize_results(self._results),
            "summary": {
                "num_successful": num_successful,
                "num_failed": num_failed,
                "num_canceled": num_canceled,
                "num_missing": len(missing_jobs),
            },
        }

    def _save_repository_info(self, registry):
        extensions = registry.list_extensions()
        extension_packages = set(["jade"])
        for ext in extensions:
            exec_module = ext[ExtensionClassType.EXECUTION].__module__
            name = exec_module.split(".")[0]
            extension_packages.add(name)

        for name in extension_packages:
            try:
                package = importlib.import_module(name)
                repo_info = RepositoryInfo(package)
                patch = os.path.join(self._output, f"{name}-diff.patch")
                repo_info.write_diff_patch(patch)
                logger.info("%s repository information: %s", name, repo_info.summary())
            except InvalidParameter:
                pass

    @staticmethod
    def _log_error_log_messages(directory):
        for event in JobSubmitter.find_error_log_messages(directory):
            log_event(event)

    @staticmethod
    def find_error_log_messages(directory):
        """Parse output log files for error messages

        Parameters
        ----------
        directory : str
            output directory

        """
        substrings = (
            "DUE TO TIME LIMIT",  # includes slurmstepd, but check this first
            "srun",
            "slurmstepd",
            "Traceback",
        )

        filenames = [os.path.join(directory, x) for x in os.listdir(directory) if x.endswith(".e")]

        if not filenames:
            return

        for line in fileinput.input(filenames):
            for substring in substrings:
                if substring in line:
                    event = StructuredLogEvent(
                        source="submitter",
                        category=EVENT_CATEGORY_ERROR,
                        name=EVENT_NAME_ERROR_LOG,
                        message="Detected error message in log.",
                        error=substring,
                        filename=fileinput.filename(),
                        line_number=fileinput.lineno(),
                        text=line.strip(),
                    )
                    yield event
                    # Only find one match in a single line.
                    break

    @staticmethod
    def generate_reports(directory, resource_monitor_type):
        """Create reports summarizing the output results of a set of jobs.

        Parameters
        ----------
        directory : str
            output directory
        resource_monitor_type : ResourceMonitorType

        """
        commands = [
            (f"jade show-results -o {directory}", "results.txt"),
            (f"jade show-events -o {directory} --categories Error", "errors.txt"),
        ]
        if resource_monitor_type != ResourceMonitorType.NONE:
            commands.append((f"jade stats show -o {directory}", "stats.txt"))
            commands.append((f"jade stats show -o {directory} -j", "stats_summary.json"))
        if resource_monitor_type == ResourceMonitorType.PERIODIC:
            commands.append((f"jade stats plot -o {directory}", None))

        reports = []
        for cmd in commands:
            output = {}
            ret = run_command(cmd[0], output=output)
            if ret != 0:
                logger.error("Failed to run [%s]: %s: %s", cmd, ret, output["stderr"])

            if cmd[1] is not None:
                filename = os.path.join(directory, cmd[1])
                with open(filename, "w") as f_out:
                    if "json" not in cmd[1]:
                        f_out.write(cmd[0] + "\n\n")
                    f_out.write(output["stdout"])
                    reports.append(filename)

        logger.info("Generated reports %s.", " ".join(reports))

        return 0

    def _submit_to_hpc(self, cluster):
        hpc_submitter = HpcSubmitter(
            self._config,
            self._config_file,
            cluster,
            self._output,
        )

        if hpc_submitter.run():
            logger.info("All submitters have completed.")
            return True

        logger.debug("jobs are still pending")
        return False

    def run_checks(self, params: SubmitterParams):
        """Checks the configuration for errors. May mutate the config."""
        self._config.check_job_dependencies(params)
        self._config.check_submission_groups(params)

    @staticmethod
    def run_submit_jobs(config_file, output, params, pipeline_stage_num=None):
        """Allows submission from an existing Python process."""
        os.makedirs(output, exist_ok=True)

        mgr = JobSubmitter.create(config_file, params, output=output)
        cluster = Cluster.create(
            output,
            mgr.config,
            pipeline_stage_num=pipeline_stage_num,
        )

        local = params.hpc_config.hpc_type == HpcType.LOCAL
        ret = 1
        try:
            status = mgr.submit_jobs(cluster, force_local=local)
            if status == Status.IN_PROGRESS:
                check_cmd = f"jade show-status -o {output}"
                if not params.dry_run:
                    print(f"Jobs are in progress. Run '{check_cmd}' for updates.")
                ret = 0
            else:
                ret = status.value
        finally:
            cluster.demote_from_submitter()
            if local:
                # These files were not used in this case.
                cluster.delete_files_internal()

        return ret



"""Common definitions for the jade package."""

import os


OUTPUT_DIR = "output"
EVENTS_DIR = "events"
JOBS_OUTPUT_DIR = "job-outputs"
SCRIPTS_DIR = "scripts"
STATS_DIR = "stats"
CONFIG_FILE = "config.json"
RESULTS_FILE = "results.json"
ANALYSIS_DIR = "analysis"
POST_PROCESSING_CONFIG_FILE = "post-config.json"
HPC_CONFIG_FILE = "hpc_config.toml"
DEFAULT_SUBMISSION_GROUP = "default"


def get_results_filename(output_dir):
    """Get the temporary results filename for all jobs.

    Parameters
    ----------
    output_dir : str
        output directory for all jobs

    Returns
    -------
    str

    """
    # This uses a CSV file because it allows for cheap appends by jobs.
    # For JSON and TOML each job would have to parse all existing jobs before
    # appending.
    return os.path.join(output_dir, "processed_results.csv")


def get_temp_results_filename(output_dir):
    """Get the temporary results filename for all jobs.

    Parameters
    ----------
    output_dir : str
        output directory for all jobs

    Returns
    -------
    str

    """
    return os.path.join(output_dir, "results.csv")



"""Defines JADE base model"""

from pydantic import BaseModel


class JadeBaseModel(BaseModel):
    """Base class for JADE models."""

    class Config:
        title = "JadeBaseModel"
        anystr_strip_whitespace = True
        validate_assignment = True
        validate_all = True
        extra = "forbid"
        use_enum_values = False



"""Synchronizes updates to the results file across jobs."""

import csv
import glob
import logging
import os

from filelock import SoftFileLock, Timeout

from jade.common import get_results_filename, get_temp_results_filename
from jade.result import Result, deserialize_result, serialize_result


logger = logging.getLogger(__name__)


class ResultsAggregator:
    """Synchronizes updates to the results file across jobs on one system."""

    def __init__(self, path, timeout=60, delimiter=","):
        """
        Constructs ResultsAggregator.

        Parameters
        ----------
        filename : str
            Full path to results filename. Must be accessible by all workers.
        timeout : int
            Lock acquistion timeout in seconds.
        delimiter : str
            Delimiter to use for CSV formatting.

        """
        self._filename = get_temp_results_filename(path)
        self._processed_filename = get_results_filename(path)
        self._lock_file = self._filename + ".lock"
        self._timeout = timeout
        self._delimiter = delimiter

    @classmethod
    def create(cls, output_dir, **kwargs):
        """Create a new instance.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        ResultsAggregator

        """
        agg = cls(output_dir, **kwargs)
        agg.create_files()
        return agg

    @classmethod
    def load(cls, output_dir, **kwargs):
        """Load an instance from an output directory.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        ResultsAggregator

        """
        return cls(output_dir, **kwargs)

    @staticmethod
    def _get_fields():
        return Result._fields

    def _do_action_under_lock(self, func, *args, **kwargs):
        # Using this instead of FileLock because it will be used across nodes
        # on the Lustre filesystem.
        lock = SoftFileLock(self._lock_file, timeout=self._timeout)
        try:
            lock.acquire(timeout=self._timeout)
        except Timeout:
            # Picked a default value such that this should not trip. If it does
            # trip under normal circumstances then we need to reconsider this.
            logger.error(
                "Failed to acquire file lock %s within %s seconds", self._lock_file, self._timeout
            )
            raise

        try:
            return func(*args, **kwargs)
        finally:
            lock.release()

    def create_files(self):
        """Initialize the results file. Should only be called by the parent
        process.

        """
        self._do_action_under_lock(self._create_files)

    def _create_files(self):
        for filename in (self._filename, self._processed_filename):
            with open(filename, "w") as f_out:
                f_out.write(self._delimiter.join(self._get_fields()))
                f_out.write("\n")

    @classmethod
    def append(cls, output_dir, result):
        """Append a result to the file.

        result : Result

        """
        aggregator = cls.load(output_dir)
        aggregator.append_result(result)

    def append_result(self, result):
        """Append a result to the file.

        result : Result

        """
        text = self._delimiter.join([str(getattr(result, x)) for x in self._get_fields()])
        self._do_action_under_lock(self._append_result, text)

    def _append_result(self, text):
        with open(self._filename, "a") as f_out:
            f_out.write(text)
            f_out.write("\n")

    def _append_processed_results(self, results):
        with open(self._processed_filename, "a") as f_out:
            for result in results:
                text = self._delimiter.join([str(getattr(result, x)) for x in self._get_fields()])
                f_out.write(text)
                f_out.write("\n")

    def _clear_temp_results(self):
        """Clear the file, once the results have been processed."""
        with open(self._filename, "w") as f_out:
            f_out.write(self._delimiter.join(self._get_fields()))
            f_out.write("\n")

    def clear_results_for_resubmission(self, jobs_to_resubmit):
        """Remove jobs that will be resubmitted from the results file.

        Parameters
        ----------
        jobs_to_resubmit : set
            Job names that will be resubmitted.

        """
        results = [x for x in self.get_results() if x.name not in jobs_to_resubmit]
        self._write_results(results)
        logger.info("Cleared %s results from %s", len(results), self._filename)

    def clear_unsuccessful_results(self):
        """Remove failed and canceled results from the results file."""
        results = [x for x in self.get_results() if x.return_code == 0]
        self._write_results(results)
        logger.info("Cleared failed results from %s", self._filename)

    def _write_results(self, results):
        _results = [serialize_result(x) for x in results]
        with open(self._processed_filename, "w") as f_out:
            writer = csv.DictWriter(f_out, fieldnames=_results[0].keys())
            writer.writeheader()
            if results:
                writer.writerows(_results)

    def get_results(self):
        """Return the current results.

        Returns
        -------
        list
            list of Result objects

        """
        return self._do_action_under_lock(self._get_all_results)

    def get_results_unsafe(self):
        """Return the results. It is up to the caller to ensure that
        a lock is not needed.

        Returns
        -------
        list
            list of Result objects

        """
        return self._get_results()

    def _get_all_results(self):
        # Include unprocessed and processed results.
        return self._get_results(processed_results=True) + self._get_results(
            processed_results=False
        )

    def _get_results(self, processed_results=True):
        filename = self._processed_filename if processed_results else self._filename
        with open(filename) as f_in:
            results = []
            reader = csv.DictReader(f_in, delimiter=self._delimiter)
            for row in reader:
                row["return_code"] = int(row["return_code"])
                row["exec_time_s"] = float(row["exec_time_s"])
                row["completion_time"] = float(row["completion_time"])
                result = deserialize_result(row)
                results.append(result)

            return results

    @classmethod
    def list_results(cls, output_dir, **kwargs):
        """Return the current results.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        list
            list of Result objects

        """
        results = cls.load(output_dir, **kwargs)
        return results.get_results()

    def process_results(self):
        """Move all temp results into the consolidated file, then clear the file.

        Returns
        -------
        list
            list of Result objects that are newly completed

        """
        return self._do_action_under_lock(self._process_results)

    def _process_results(self):
        results = self._get_results(processed_results=False)
        self._append_processed_results(results)
        self._clear_temp_results()
        return results



"""Implements the JobParametersInterface for generic_command."""

import logging
from collections import namedtuple
from typing import Dict, List, Optional, Set

from pydantic import Field, validator

from jade.models import JadeBaseModel
from jade.common import DEFAULT_SUBMISSION_GROUP
from jade.jobs.job_parameters_interface import JobParametersInterface


logger = logging.getLogger(__name__)

_EXTENSION = "generic_command"


class GenericCommandParameters(JobParametersInterface):
    """A class used for creating a job for a generic command."""

    parameters_type = namedtuple("GenericCommand", "command")

    def __init__(self, **kwargs):
        self._model = GenericCommandParametersModel(**kwargs)

    def __str__(self):
        return "<GenericCommandParameters: {}>".format(self.name)

    def __getattr__(self, name):
        if name in GenericCommandParametersModel.__fields__:
            return getattr(self._model, name)
        raise AttributeError(f"'GenericCommandParameters' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name == "extension":
            raise AttributeError(f"'GenericCommandParameters' does not allow setting 'extension'")
        if name in GenericCommandParametersModel.__fields__:
            setattr(self._model, name, value)
        self.__dict__[name] = value

    @property
    def command(self):
        if self._model.use_multi_node_manager:
            return f"jade-internal run-multi-node-job {self.name} {self._model.command}"
        return self._model.command

    @property
    def estimated_run_minutes(self):
        return self._model.estimated_run_minutes

    @property
    def extension(self):
        return _EXTENSION

    @property
    def name(self):
        return self._create_name()

    def _create_name(self):
        return str(self._model.job_id)

    def serialize(self):
        assert self._model.job_id is not None
        # If job sizes get huge then we should exclude parameters with default values.
        return self._model.dict()

    @classmethod
    def deserialize(cls, data):
        return cls(**data)

    @property
    def cancel_on_blocking_job_failure(self):
        return self._model.cancel_on_blocking_job_failure

    def get_blocking_jobs(self):
        return self._model.blocked_by

    def remove_blocking_job(self, name):
        self._model.blocked_by.remove(name)

    def set_blocking_jobs(self, blocking_jobs):
        self._model.blocked_by = blocking_jobs

    @property
    def submission_group(self):
        return self._model.submission_group


class GenericCommandParametersModel(JadeBaseModel):
    """Model definition for generic command parameters"""

    use_multi_node_manager: Optional[bool] = Field(
        title="use_multi_node_manager",
        description="If true JADE will run this job with its multi-node manager.",
        default=False,
    )
    command: str = Field(
        title="command",
        description="Command that can be invoked in a terminal (shell characters not allowed)",
    )
    blocked_by: Optional[Set[str]] = Field(
        title="blocked_by",
        description="Array of job names that must complete before this one can start.",
        default=set(),
    )
    cancel_on_blocking_job_failure: Optional[bool] = Field(
        title="cancel_on_blocking_job_failure",
        description="If true JADE will cancel this job if any of its blocking jobs fail.",
        default=False,
    )
    estimated_run_minutes: Optional[int] = Field(
        title="estimated_run_minutes",
        description="JADE will use this value along with num-processes and walltime to "
        "build per-node batches of jobs if time-based-batching is enabled.",
    )
    submission_group: Optional[str] = Field(
        title="submission_group",
        description="Optional name of a submission group",
        default=DEFAULT_SUBMISSION_GROUP,
    )
    append_output_dir: Optional[bool] = Field(
        title="append_output_dir",
        description="If true JADE will append --jade-runtime-output=X where X is the output "
        "directory specified in jade submit-jobs.",
        default=False,
    )
    ext: Optional[Dict] = Field(
        title="ext",
        description="User-defined extension data to be used at runtime. Must be serializable in "
        "JSON format.",
        default={},
    )
    job_id: Optional[int] = Field(
        title="job_id",
        description="Unique job identifier, generated by JADE",
    )
    extension: Optional[str] = Field(
        title="extension",
        description="job extension type, generated by JADE",
        default=_EXTENSION,
    )

    @validator("append_output_dir")
    def handle_append_output_dir(cls, value, values):
        if values["use_multi_node_manager"]:
            logger.debug("Override 'append_output_dir' because 'use_multi_node_manager' is set")
            return True
        return value

    @validator("blocked_by")
    def handle_blocked_by(cls, value):
        return {str(x) for x in value}



"""CLI to display and manage config files."""

import json
import logging
import os
import re
import sys
import tempfile
from pathlib import Path

import click
from prettytable import PrettyTable

from jade.cli.common import COMMON_SUBMITTER_OPTIONS, add_options, make_submitter_params
from jade.common import CONFIG_FILE, HPC_CONFIG_FILE
from jade.extensions.generic_command import GenericCommandConfiguration, GenericCommandParameters
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.utils.utils import dump_data, load_data
from jade.models import HpcConfig, SlurmConfig, FakeHpcConfig, LocalHpcConfig


logger = logging.getLogger(__name__)


@click.group()
def config():
    """Manage a JADE configuration."""
    setup_logging("config", None)


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "-c",
    "--config-file",
    default=CONFIG_FILE,
    show_default=True,
    help="config file to generate.",
)
@click.option(
    "-C",
    "--cancel-on-blocking-job-failure",
    is_flag=True,
    default=False,
    show_default=True,
    help="cancel any job if one of its blocking jobs fails.",
)
@click.option(
    "-m",
    "--minutes-per-job",
    type=int,
    help="estimated minutes per job.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable verbose log output.",
)
def create(filename, config_file, cancel_on_blocking_job_failure, minutes_per_job, verbose):
    """Create a config file from a filename with a list of executable commands."""
    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("auto_config", None, console_level=level)

    config = GenericCommandConfiguration.auto_config(
        filename,
        cancel_on_blocking_job_failure=cancel_on_blocking_job_failure,
        minutes_per_job=minutes_per_job,
    )
    print(f"Created configuration with {config.get_num_jobs()} jobs.")
    config.dump(config_file)
    print(f"Dumped configuration to {config_file}.\n")


@click.command()
@click.option(
    "-a",
    "--account",
    default="",
    help="HPC account/allocation",
)
@click.option(
    "-c",
    "--config-file",
    default="hpc_config.toml",
    show_default=True,
    help="config file to create",
)
@click.option(
    "-m",
    "--mem",
    default=None,
    help="Amount of memory required by a single node.",
)
@click.option(
    "-p",
    "--partition",
    default=None,
    help="HPC partition",
)
@click.option(
    "-q",
    "--qos",
    default=None,
    type=str,
    help="QoS value",
)
@click.option(
    "-t",
    "--hpc-type",
    type=click.Choice(["slurm", "fake", "local"]),
    default="slurm",
    show_default=True,
    help="HPC queueing system",
)
@click.option(
    "--tmp",
    default=None,
    help="Amount of local storage space required by a single node.",
)
@click.option(
    "-w",
    "--walltime",
    default="4:00:00",
    help="HPC walltime",
)
def hpc(account, config_file, mem, partition, qos, hpc_type, tmp, walltime):
    """Create an HPC config file."""
    if hpc_type == "slurm":
        hpc = SlurmConfig(
            account=account,
            mem=mem,
            partition=partition,
            qos=qos,
            tmp=tmp,
            walltime=walltime,
        )
    elif hpc_type == "fake":
        hpc = FakeHpcConfig(walltime=walltime)
    else:
        assert hpc_type == "local"
        hpc = LocalHpcConfig()

    # This converts enums to values.
    data = json.loads(HpcConfig(hpc_type=hpc_type, hpc=hpc).json())
    dump_data(data, config_file)
    print(f"Created HPC config file {config_file}")


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
@click.option(
    "-f",
    "--fields",
    multiple=True,
    help="include in output table; can specify mulitple times",
)
@click.option(
    "-n",
    "--no-blocked-by",
    is_flag=True,
    default=False,
    show_default=True,
    help="exclude blocking jobs in table",
)
def show(config_file, fields, no_blocked_by):
    """Show the jobs in the configuration."""
    _show(config_file, fields, not no_blocked_by)


# This is a standalone function so that it can be called from _filter.
def _show(config_file, fields=None, blocked_by=True):
    config = create_config_from_file(config_file)
    num_jobs = config.get_num_jobs()
    print(f"Num jobs: {num_jobs}")
    if num_jobs == 0:
        return

    # generic_command jobs have a command field which is very useful.
    # Other extensions do not.
    has_command = False
    for job in config.iter_jobs():
        if isinstance(job, GenericCommandParameters):
            has_command = True
            break

    field_names = ["index", "name"]
    if has_command:
        field_names.append("command")
    if blocked_by:
        field_names.append("blocked_by (job names)")
    if fields is not None:
        field_names += fields

    table = PrettyTable()
    table.field_names = field_names
    for i, job in enumerate(config.iter_jobs()):
        job_dict = job.serialize()
        row = [i, job.name]
        if has_command:
            row.append(job_dict.get("command", ""))
        if blocked_by:
            blocking_jobs = sorted(list(job.get_blocking_jobs()))
            text = ", ".join(blocking_jobs)
            if len(text) > 50:
                text = f"truncated...blocked by {len(blocking_jobs)} jobs"
            row.append(text)
        if fields is not None:
            for field in fields:
                row.append(job_dict.get(field, ""))
        table.add_row(row)
    print(table)


@click.command("filter")
@click.argument("config_file", type=click.Path(exists=True))
@click.argument("indices", nargs=-1)
@click.option(
    "-o",
    "--output-file",
    help="Create new config file with filtered jobs.",
)
@click.option(
    "-f",
    "--fields",
    type=str,
    multiple=True,
    nargs=2,
    help="Filter on field value. Multiple accepted.",
)
@click.option(
    "-s",
    "--show-config",
    is_flag=True,
    show_default=True,
    default=False,
    help="Show the new config (only applicable if output-file is provided).",
)
# Named _filter to avoid collisions with the built-in function.
def _filter(config_file, output_file, indices, fields, show_config=False):
    """Filters jobs in CONFIG_FILE. Prints the new jobs to the console or
    optionally creates a new file.

    Note: This does not detect duplicate ranges.

    \b
    Examples:
    1. Select the first job. Output only.
       jade config filter c1.json 0
    2. Select indices 0-4, 10-14, 20, 25, create new file.
       jade config filter c1.json :5 10:15 20 25 -o c2.json
    3. Select the last 5 jobs. Note the use of '--' to prevent '-5' from being
       treated as an option.
       jade config filter c1.json -o c2.json -- -5:
    4. Select indices 5 through the end.
       jade config filter c1.json -o c2.json 5:
    5. Select jobs with parameters param1=green and param2=3.
       jade config filter c1.json -o c2.json -f param1 green -f param2 3

    """
    cfg = load_data(config_file)
    jobs = cfg["jobs"]
    if not jobs:
        print("The configuration has no jobs", file=sys.stderr)
        sys.exit(1)

    if output_file is None:
        handle, new_config_file = tempfile.mkstemp(suffix=".json")
        os.close(handle)
        show_config = True
    else:
        new_config_file = output_file

    try:
        if not new_config_file.endswith(".json"):
            print("new_config_file must have extension .json", file=sys.stderr)
            sys.exit(1)

        orig_len = len(jobs)
        new_jobs = []
        regex_int = re.compile(r"^(?P<index>\d+)$")
        regex_range = re.compile(r"^(?P<start>[\d-]*):(?P<end>[\d-]*)$")
        for index in indices:
            match = regex_int.search(index)
            if match:
                i = int(match.groupdict()["index"])
                new_jobs.append(jobs[i])
                continue
            match = regex_range.search(index)
            if match:
                start = match.groupdict()["start"]
                if start == "":
                    start = None
                else:
                    start = int(start)
                end = match.groupdict()["end"]
                if end == "":
                    end = None
                else:
                    end = int(end)
                new_jobs += jobs[start:end]

        # Note: when looking at just the JSON, there is no way to get the job name,
        # and so we can't check for duplicates.

        if not new_jobs:
            new_jobs = jobs

        if fields:
            final_jobs = []
            for job in new_jobs:
                matched = True
                for field in fields:
                    if str(job[field[0]]) != field[1]:
                        matched = False
                        break
                if matched:
                    final_jobs.append(job)

            new_jobs = final_jobs

        cfg["jobs"] = new_jobs
        new_len = len(cfg["jobs"])
        dump_data(cfg, new_config_file, indent=4)
        print(f"Filtered {config_file} ({orig_len} jobs) into ({new_len} jobs)\n")
        if output_file is not None:
            print(f"Wrote new config to {output_file}")

        if show_config:
            _show(new_config_file)
    finally:
        if output_file is None:
            os.remove(new_config_file)


@click.command()
@click.option(
    "-c",
    "--config-file",
    default="submitter_params.json",
    show_default=True,
    type=Path,
    help="config file to create; can be .toml or .json",
)
@add_options(COMMON_SUBMITTER_OPTIONS)
def submitter_params(
    config_file=None,
    dry_run=None,
    per_node_batch_size=None,
    hpc_config=None,
    local=None,
    max_nodes=None,
    poll_interval=None,
    resource_monitor_interval=None,
    resource_monitor_type=None,
    num_processes=None,
    verbose=None,
    reports=None,
    enable_singularity=None,
    container=None,
    try_add_blocked_jobs=None,
    time_based_batching=None,
    node_setup_script=None,
    node_shutdown_script=None,
    no_distributed_submitter=None,
):
    """Create parameters for use in 'jade submit-jobs'."""
    params = make_submitter_params(
        per_node_batch_size=per_node_batch_size,
        dry_run=dry_run,
        hpc_config=hpc_config,
        local=local,
        max_nodes=max_nodes,
        poll_interval=poll_interval,
        resource_monitor_interval=resource_monitor_interval,
        resource_monitor_type=resource_monitor_type,
        num_processes=num_processes,
        verbose=verbose,
        reports=reports,
        enable_singularity=enable_singularity,
        container=container,
        try_add_blocked_jobs=try_add_blocked_jobs,
        time_based_batching=time_based_batching,
        node_setup_script=node_setup_script,
        node_shutdown_script=node_shutdown_script,
        no_distributed_submitter=no_distributed_submitter,
    )
    # This converts enums to values.
    data = json.loads(params.json())
    if config_file.suffix == ".json":
        dump_data(data, config_file, indent=2)
    else:
        dump_data(data, config_file)
    print(f"Created submitter parameter file {config_file}")


config.add_command(create)
config.add_command(hpc)
config.add_command(show)
config.add_command(_filter)
config.add_command(submitter_params)
